#!/usr/bin/env python

import urllib2
import urlparse
from xml.dom import minidom
from HTMLParser import HTMLParser
import sys
from optparse import OptionParser
import urllib

admin_url = 'http://%s/admin/advanced'
post_url = 'http://%s/admin/asipura.spa'

def getConfigurationFromDevice(url, username, pw, realm):
    authhandler = urllib2.HTTPDigestAuthHandler()
    authhandler.add_password(realm, url, username, pw)
    opener = urllib2.build_opener(authhandler)
    urllib2.install_opener(opener)
    pagehandle = urllib2.urlopen(url)
    return pagehandle.read()

def writeConfigurationToDevice(url, username, pw, realm, configValues):
    authhandler = urllib2.HTTPDigestAuthHandler()
    authhandler.add_password(realm, url, username, pw)
    opener = urllib2.build_opener(authhandler)
    urllib2.install_opener(opener)
    pagehandle = urllib2.urlopen(url, configValues)
    return pagehandle.read()

class MyHTMLParser(HTMLParser):
    """
    Parse the field names, numbers and values out of the sipura's HTML file. 
    This is fairly straightforward because the code looks like:

    <tr bgcolor="#d3d3d3"><td>Enable Web Server:<td><select class="inputw" name="31215"><option selected value="1">yes</option><option  value="0">no</option></select><td>Web Server Port:<td><input class="inputw" size="20" name="25710" value="80" maxlength=2047>

    So we look for table cells and the following input/select fields.
    """

    def __init__(self):
        HTMLParser.__init__(self)
        self._startedCell = False
        self._startedOption = False
        self._fieldNameToNum = {}
        self._configValues = []
        self._possibleValues = {}
        self._select = None
        self._configValuesByTab = {}

    def findAttr(self, attrs, attr):
        for a, val in attrs:
            if a == attr:
                return val
        return None

    def hasAttr(self, attrs, attr):
        for a, val in attrs:
            if a == attr:
                return True
        return False

    def fixFieldName(self, fieldname):
        # replace characters with underscores to ensure the resulting field name is 
        # valid as a python identifier
        fieldname = fieldname.lower()
        fieldname = fieldname.replace(' ', '_')
        fieldname = fieldname.replace('-', '_')
        fieldname = fieldname.replace('(', '_')
        fieldname = fieldname.replace(')', '_')
        fieldname = fieldname.replace('/', '_')
        return fieldname

    def uniqueFieldName(self, fieldname):
        name = self._tab + "_" + self.fixFieldName(fieldname)
        assert not self._fieldNameToNum.has_key(fieldname), "%s, %s" % (fieldname, attrs)
        return name

    def handle_starttag(self, tag, attrs):
        if tag == 'td':
            self._startedCell = True
        else:
            self._startedCell = False
        if tag == 'div':
            if self.findAttr(attrs, 'class') == 'tab-page':
                self._tab = self.fixFieldName(self.findAttr(attrs, 'id'))
                self._configValuesByTab[self._tab] = []
        if tag == 'input':
            if self.findAttr(attrs, 'type') in ['submit', 'reset']:
                return  # don't do anything
            fieldname = self.uniqueFieldName(self._fieldName)
            self._fieldNameToNum[fieldname] = self.findAttr(attrs, 'name')
            value = self.findAttr(attrs, 'value')
            self._configValuesByTab[self._tab].append((fieldname, value))
        elif tag == 'select':
            fieldname = self.uniqueFieldName(self._fieldName)
            self._fieldNameToNum[fieldname] = self.findAttr(attrs, 'name')
            self._select = fieldname
            self._options = []
            self._selectedOptionValue = None
        elif tag == 'option':
            self._startedOption = True
            value = self.findAttr(attrs, 'value')
            self._curOptionValue = value
            if self.hasAttr(attrs, 'selected'):
                self._selectedOptionValue = value

    def handle_endtag(self, tag):
        self._startedCell = self._startedOption = False
        if tag == 'select':
            self._configValuesByTab[self._tab].append((self._select, self._selectedOptionValue))
            self._possibleValues[self._select] = self._options
        elif tag == 'option':
            self._options.append((self._curOptionValue, self._curOptionDesc))

    def handle_data(self, data):
        if self._startedCell and data[-1] == ':':
            self._fieldName = data[:-1]
        elif self._startedOption:
            self._curOptionDesc = data

    def fieldNameMap(self):
        return self._fieldNameToNum.copy()

    def configValues(self):
        values = []
        for tab, configValues in self._configValuesByTab.items():
            values.extend(configValues)
        return values

    def configValuesByTab(self):
        return self._configValuesByTab.copy()

    def possibleValues(self):
        return self._possibleValues.copy()


def configValuesToFile(configValuesByTab, possibleValues):
    lines = []
    for tab, configValues in configValuesByTab.items():
        if configValues:
            lines.append('''
#
# %s
#       ''' % tab)
        for fieldname, value in configValues:
            if possibleValues.has_key(fieldname):
                valuestring = ", ".join(map(lambda opt: "'%s'" % opt[1], possibleValues[fieldname]))
                comment = '# possible values: %s\n' % (valuestring)

                # find the description for the current value
                for val, desc in possibleValues[fieldname]:
                    if value == val:
                        value = desc
                        break
            else:
                comment = ''
            line = "%s%s = '%s'" % (comment, fieldname, value)
            lines.append(line)
    return "\n\n".join(lines) + "\n"

def configFileToDict(input_file, fieldNameMap, possibleValues):
    """
    Read configuration file into a dict.
    """
    config = {}
    configlines = file(input_file).readlines()
    for line in configlines:
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        i = line.find('=')
        if i == -1:
            continue
        key, value = line.split('=')
        key = key.strip()
        value = value.strip()
        if value[0] == "'" and value[-1] == "'":
            value = value[1:-1]
        config[key] = value
        
    mappedConfigValues = {}
    for key, value in config.items():
        if key == '__builtins__':
            continue
        if key.find('password') and value.startswith('****'):
            print "Ignoring setting: %s = '%s'" % (key, value)
            continue
        if possibleValues.has_key(key):
            for val, desc in possibleValues[key]:
                if value == desc:
                    value = val
        newkey = fieldNameMap[key]
        mappedConfigValues[newkey] = value
    return mappedConfigValues

def parseHost(host):
    i = host.rfind('@')
    if i != -1:
        usernamePassword, hostPort = host[:i], host[i+1:]
        i = usernamePassword.find(':')
        if i != -1:
            username, password = usernamePassword[:i], usernamePassword[i+1:]
    else:
        username = password = ''
        hostPort = host
    i = hostPort.find(':')
    if i != -1:
        host, port = hostPort[:i], hostPort[i+1:]
    else:
        host = hostPort
        port = ''
    return username, password, host, port

if __name__ == "__main__":
    usage = '''
%prog <src> [<dest>]
    
    %prog can read the active configuration from a Sipura SPA device and 
    store it in a text file, or can take an existing text file and write 
    that configuration to the sipura device.

    One of <src> or <dest> should be a local file path, and the other 
    should be a URL to your sipura device, with the username and password 
    embedded into the URL.

    Examples:

    Write sipura config to standard output:
        %prog http://admin:mypassword@mysipura/

    Copy sipura config into a config file:
        %prog http://admin:mypassword@mysipura/ sipura.config

    Or use an explicit IP if necessary:
        %prog http://admin:mypassword@192.168.0.5/ sipura.config

    Copy a config file to the sipura device:
        %prog sipura.config http://admin:mypassword@192.168.0.5/

    Notes:
        When reading a config from a sipura device, this program 
        unfortunately cannot retrieve the passwords.  They will appear 
        as stars ("****").  
        
        It is however safe to write such a configuration back to the 
        device; such passwords will be ignored and the passwords in the 
        device will be left unchanged.

        This program has been tested with the Sipura SPA 3000, software
        version 2.0.13g.  Your mileage may vary; please report
        problems and successes.
    '''
    optparser = OptionParser(usage=usage)
    (options, args) = optparser.parse_args()
    urls = 0
    for arg in args:
        if arg.startswith('http://'):
            urls += 1
    if urls == 0:
        optparser.error("Either the source or the destination must be a URL.")
    elif urls >= 2:
        optparser.error("Only the source OR the destination should be a URL; the other should be a local file.")
    if args[0].startswith('http://'):
        url = args[0]
        action = 'read-config'
        if len(args) >= 2:
            options.output_file = args[1]
        else:
            options.output_file = '-'   # output to stdout
    else:
        if len(args) < 2:
            optparser.error("If the source is a file, you must specify the destination (the URL to a sipura device).")
        url = args[1]
        action = 'write-config'
        options.input_file = args[0]

    # break apart url
    scheme, host, uri = urlparse.urlparse(url)[:3]
    username, password, host, port = parseHost(host)
    if port:
        hostPort = '%s:%s' % (host, port)
    else:
        hostPort = host

    if not username or not password:
        optparser.error("You need to specify the username and password for the sipura device.")

    try:
        if action == 'read-config':
            url = admin_url % hostPort
            html = getConfigurationFromDevice(url, username, password, 'spa admin')
            htmlparser = MyHTMLParser()
            htmlparser.feed(html)
            if options.output_file != '-':
                out = file(options.output_file, 'w')
            else:
                out = sys.stdout
            out.write(configValuesToFile(htmlparser.configValuesByTab(), htmlparser.possibleValues()))
        elif action == 'write-config':

            # grab output from sipura so that we can get the fieldname->number mapping
            url = admin_url % hostPort
            html = getConfigurationFromDevice(url, username, password, 'spa admin')
            htmlparser = MyHTMLParser()
            htmlparser.feed(html)
            mapping = htmlparser.fieldNameMap()
            possibleValues = htmlparser.possibleValues()

            # read the config file
            params = configFileToDict(options.input_file, mapping, possibleValues)
            
            params = urllib.urlencode(params)
            url = post_url % hostPort
            writeConfigurationToDevice(url, username, password, 'spa admin', params)
            print 'Wrote configuration to device.'
    except urllib2.HTTPError, e:
        print 'HTTP Error %d (%s) occurred when communicating with sipura device.' % (e.code, e.msg)
        if e.code == 401:
            print 'Check that you have the correct username and password.'
    except urllib2.URLError, e:
        if hasattr(e, 'reason'):
            print 'Error %d (%s) occurred when communicating with sipura device.' % (e.reason[0], e.reason[1])
            if e.reason[0] == 111:
                print 'Check that you have the correct hostname and port number.'
        else:
            print str(e)
